/*
 * Bundesliga is Copyright (c) 1997-2001 by Thomas Pundt
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose, without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice and this paragraph and the
 * following two paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 * THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
 * BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include <PalmOS.h>
#include <ExgMgr.h>

#include "bl.h"
#include "blrcp.h"

/**
 * Database Handling Routinen
 */

DmOpenRef db;
static SelectedDBInfo dbInfo[20];
char* Items[20] = { 0, };
static Err err = 0;
Int16 nrOfItems = 0;

static Int16 DBNameNr(char*) DBFUNC;
static UInt32 dbhash(char* name) DBFUNC;
static Boolean isDB(UInt16 cardNo, LocalID dbID) DBFUNC;
static Err WriteDBData(const void *, UInt32 *, void *) DBFUNC;
static Int16 dbSortFunc(void* a, void* b, Int32 other) DBFUNC;

/*
 * Create a new Database, install all records.
 */
DmOpenRef CreateDatabase(char* name)
{
  UInt16      index;
  DmOpenRef db;
  LocalID   dbID;
  MemHandle  p;

  if (DBNameNr(name)) {
    DisplayErrorMessage(ID_ErrorMessages+2);
    err = -1;
    return NULL;
  }

  if ( (err = DmCreateDatabase(0, name, 'tpBL', 'Data', false)) )
    return NULL;
  dbID = DmFindDatabase(0, name);
  db = DmOpenDatabase(0, dbID, dmModeReadWrite);

  index = 0;
  p = MemHandleLock(DmNewRecord(db, &index, sizeof(prefs)));
  DmWrite(p, 0, &prefs, sizeof(prefs));
  MemPtrUnlock(p);
  DmReleaseRecord(db, index, true);

  return db;
}

/*
 * Read and store contents of Database db "locally".
 */
Err ReadDatabase(DmOpenRef db)
{
  char *p, *q;
  UInt16 i;
  
  p = MemHandleLock(DmQueryRecord(db,0));
  MemMove(&prefs, p, sizeof(prefs));
  MemPtrUnlock(p);

  if (prefs.Version < 6)
    return (appErrorClass|tooOld);
  if (prefs.Version < 7) {
    prefs.favTeam = 0;
    prefs.Version = 7;
  }

  if (MStatus) {
    i = MemHandleResize(MStatusHandle, prefs.NumTeams*sizeof(MannStatus));
  } else {
    MStatusHandle = MemHandleNew(prefs.NumTeams*sizeof(MannStatus));
  }
  MStatus = MemHandleLock(MStatusHandle);
  p = q = (char*)MemHandleLock(DmQueryRecord(db,1));
  for (i=0; i<prefs.NumTeams; i++) {
    MStatus[i].team = p;
    p += StrLen(p)+1;
  }
  MemPtrUnlock(q);
  MemHandleUnlock(MStatusHandle);

  return 0;
}

static Boolean isDB(UInt16 cardNo, LocalID dbID)
{
  Int16 i;
  if (nrOfItems==0) 
    FindInstalledDatabases();
  for (i=0; i<nrOfItems; i++) {
    if ( (dbInfo[i].cardNo == cardNo) &&
         (dbInfo[i].dbID == dbID) )
      return true;
  }
  return false;
}
 

/*
 * Trys to open database dbID.
 */
Err OpenDatabase(UInt16 cardNo, LocalID dbID)
{
  db = 0;

  if (isDB(cardNo,dbID))
    db = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);
  else
    dbID = 0;

  if (nrOfItems == 0) 
    return (appErrorClass|noDatabaseInstalled);

  if (dbID == 0) {
    cardNo = dbInfo[0].cardNo;
    dbID   = dbInfo[0].dbID;
    db = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);
  }

  if (db) {
    if ( ReadDatabase(db) ) {
      DmCloseDatabase(db);
      db = 0;
      return (appErrorClass|tooOld);
    }
  } else {
    return DmGetLastErr();
  }

  if (!prefs.NumMatchDays)
    prefs.NumMatchDays = 2*(prefs.NumTeams + (prefs.NumTeams%2) -1);

  // security check; database might be script generated and contain
  // invalid values.
  if (prefs.matchDay<1 || prefs.matchDay+2 > DmNumRecords(db))
    prefs.matchDay = 1;

  return 0;
}

void SavePreferences(void)
{
  void *p = MemHandleLock(DmResizeRecord(db, 0, sizeof(prefs)));
  DmWrite(p, 0, &prefs, sizeof(prefs));
  MemPtrUnlock(p);
  DmReleaseRecord(db, 0, true);
}

/*
 * Saves program status to open database and updates database
 * "nr" attributes.
 */
void SaveStatus(Int16 nr)
{
  Int16 i;
  UInt16 size;
  
  prefs.matchDay = matchDay;

  if (!db)
    return;

  SavePreferences();

  dbInfo[nr].dbAttr |= dmHdrAttrBackup;
  DmSetDatabaseInfo (dbInfo[nr].cardNo, dbInfo[nr].dbID, NULL, 
		     &dbInfo[nr].dbAttr, 
		     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
  DmCloseDatabase(db);

  if (appPrefs.options & dontSaveEM)
    appPrefs.options &= ~editMode;

  /* remember sort order the user chose for the leagues */
  for (i = 0; i<20; i++) {
    if (i<nrOfItems) {
      appPrefs.dbHash[i] = dbhash(Items[i]);
    } else {
      appPrefs.dbHash[i] = 0;
    }
  }

  appPrefs.Version = 1;
  size = sizeof(appPrefs);
  PrefSetAppPreferences('tpBL', 0, appPrefs.Version, &appPrefs, size, false);
}

void InsertNewRecord(DmOpenRef db, UInt16 *index, UInt16 numMatches)
{
  char *p;
  matchType m;
  UInt16 j = numMatches<<8;

  p = MemHandleLock(DmNewRecord(db, index, 8*numMatches+2));
  DmWrite(p,0,&j,2);
  for (j=0; j<numMatches; j++) {
    m.m1 = j*2+1; m.m2=j*2+2; m.e1=255; m.e2=255; m.Datum=0;
    DmWrite(p,2+sizeof(matchType)*j,&m,sizeof(matchType));
  }
  MemPtrUnlock(p);
  DmReleaseRecord(db, *index, true);
}

/*
 * write "sane" values to freshly created database db
 */
void InitNewDatabase(DmOpenRef db, BlDatabasePreferencesType* newPrefs)
{
  char *p, buf[8];
  UInt16 i, index;

  index = 0;
  p = MemHandleLock(DmNewRecord(db, &index, sizeof(prefs)));
  DmWrite(p, 0, newPrefs, sizeof(prefs));
  MemPtrUnlock(p);
  DmReleaseRecord(db, 0, true);

  StrCopy(buf,"Team A");
  index = 1;
  p = MemHandleLock(DmNewRecord(db, &index, 7*newPrefs->NumTeams));
  for (i=0; i<newPrefs->NumTeams; i++) {
    buf[5] = 'A'+i;
    DmWrite(p,i*7,buf,7);
  }
  MemPtrUnlock(p);
  DmReleaseRecord(db, 1, true);

  for (i=0; i<newPrefs->NumMatchDays; i++) {
    index = i+2;
    InsertNewRecord(db, &index, newPrefs->NumTeams>>1);
  }

  nrOfItems = 0;
}

/*
 * looks for installed databases and stores result into global dbInfo[].
 */
Int16 FindInstalledDatabases(void)
{
  void*   RecPointer;
  DmOpenRef db;
  BlDatabasePreferencesType newPrefs;
  Boolean newSearch = true;
  DmSearchStateType stateInfo;
  Int16 i = 0;

  while (DmGetNextDatabaseByTypeCreator(newSearch, &stateInfo, 
	 'Data', 'tpBL', false, &dbInfo[i].cardNo, &dbInfo[i].dbID) == 0 
	 && i<20) {
    newSearch = false;

    db = DmOpenDatabase(dbInfo[i].cardNo, dbInfo[i].dbID, dmModeReadOnly);
    RecPointer = MemHandleLock(DmQueryRecord(db, 0));
    MemMove(&newPrefs, RecPointer, sizeof(newPrefs));
    MemPtrUnlock(RecPointer);
    DmCloseDatabase(db);
  
    if (newPrefs.Version < 6) {
      continue;
    }
    i++;
  }
  nrOfItems = i;
  return nrOfItems;
}  

/*
 * returns number of database with name "name"
 */
static Int16 DBNameNr(char* name)
{
  Int16 i;
  for (i=0; Items[i]; i++) {
    if (!StrCompare(name,Items[i]))
      return i+1;
  }
  return 0;
}

/* Sorting function for use with SysInsertionSort */
static Int16 dbSortFunc(void* a, void* b, Int32 other)
{
  UInt32 keyA = dbhash(((SelectedDBInfo*)a)->dbName);
  UInt32 keyB = dbhash(((SelectedDBInfo*)b)->dbName);
  Int16 i;

  for (i=0; i<20; i++) {
    if (keyA == appPrefs.dbHash[i])
      return -1;
    if (keyB == appPrefs.dbHash[i])
      return 1;
  }
  return 0;
}

void ChooseDB(void)
{
  Int16 i;
  ListPtr p;

  if (nrOfItems == 0)
    FindInstalledDatabases();

  for (i=0; i<nrOfItems; i++) {
    DmDatabaseInfo(dbInfo[i].cardNo, 
		   dbInfo[i].dbID, 
		   dbInfo[i].dbName, 
		   NULL, NULL, NULL, NULL, NULL, 
		   NULL, NULL, NULL, NULL, NULL);
  }

  /* preserve choosen sort order */
  SysInsertionSort(dbInfo,nrOfItems,sizeof(SelectedDBInfo),dbSortFunc,0);

  for (i=0; i<nrOfItems; i++) {
    Items[i] = i<nrOfItems?dbInfo[i].dbName:0;

    /* this must appear _after_ sorting the items. Having it here
     * doesn't hurt, because appPrefs.dbID is consistently updated
     */
    if ( (appPrefs.dbID == dbInfo[i].dbID) && 
	 (appPrefs.cardNo == dbInfo[i].cardNo) ) 
	 DBNr = i;
  }

  p = FrmGetListPtr(ID_FrmBl, ID_FrmPopup+1);
  LstSetListChoices(p, Items, nrOfItems);
  LstSetHeight(p, nrOfItems>10?10:nrOfItems);
}

void SelectedDB(Int16 i, SelectedDBInfo *info)
{
  MemMove(info, &dbInfo[i], sizeof(SelectedDBInfo));
  appPrefs.dbID = dbInfo[i].dbID;
  appPrefs.cardNo = dbInfo[i].cardNo;
}

Err DeleteDatabase(Int16 i)
{
  Err err;
  Int16 j;
  err = DmCloseDatabase(db);
  if ((err = DmDeleteDatabase(dbInfo[i].cardNo, dbInfo[i].dbID))) {
    DisplayErrorMessage(ID_ErrorMessages+3);
    return err;
  }
  for (j=i; j<nrOfItems-1; j++) {
    MemMove(&dbInfo[j], &dbInfo[j+1], sizeof(dbInfo[j]));
  }
  nrOfItems--;
  if (nrOfItems == 0) {
    db = 0;
    return (appErrorClass|noDatabaseInstalled);
  }
  ChooseDB();
  return 0;
}

Boolean RenameDatabase(Int16 i, char* name)
{
  if (DmSetDatabaseInfo (dbInfo[i].cardNo, dbInfo[i].dbID, 
			 name, NULL, NULL, NULL, NULL, NULL, 
			 NULL, NULL, NULL, NULL, NULL) == 0) {
    StrCopy(dbInfo[i].dbName,name);
    return true;
  }
  return false;
}

/* 
 * Hash function that generates keys from database names. The keys rather
 * than the database names are stored in the application preferences.
 */
static UInt32 dbhash(char* name)
{
  UInt32 key = 0;
  Int8 c;

  while ((c = *name++)) {
    key += c;
    key ^= (key << 24) | (key << 16) | (key << 8);
  }

  return key;
}

/*
 * exchange two dbInfo entries
 */
void swapDbInfo(Int16 i, Int16 j)
{
  SelectedDBInfo tmp = dbInfo[i];
  dbInfo[i] = dbInfo[j];
  dbInfo[j] = tmp;
}

/*
 * Beaming support; we transfer a complete database and let 
 * the launcher application do the dirty work :-)
 */
static Err WriteDBData(const void *dataP, UInt32 *sizeP, void *userDataP)
{
  Err err;
  *sizeP = ExgSend((ExgSocketType*)userDataP,(void*)dataP,*sizeP,&err);
  return err;
}

void BeamDatabase(Int16 no)
{
  ExgSocketType s;
  char name[36];
  Err err;

  if (!FeatureSet30) {
    DisplayInformationMessage(ID_InformationMessages+1);
    return;
  }
  
  StrCopy(name, dbInfo[no].dbName);
  StrCat(name, ".pdb");

  MemSet(&s, sizeof(s), 0);
  s.description = name;
  s.name = name;

  if (! (err = ExgPut(&s)) ) {
    err = ExgDBWrite(WriteDBData,&s,NULL,dbInfo[no].dbID,dbInfo[no].cardNo);
  }
  ExgDisconnect(&s, err);

}

